import { goToOauthPage, reconstructProjectPlan, scheduleUnresolved, appendSchedule, isFailure } from '@fschopp/project-planning-for-you-track';
import { AlertsCtrl } from '../alerts/alerts-ctrl.js';
import S from 's-js';
import { deepEquals } from '../../utils/deep-equals.js';
import { unreachableCase } from '../../utils/typescript.js';
import { SettingsCtrl } from '../settings/settings-ctrl.js';
import { toNormalizedPlainSettings } from '../settings/settings-model.js';
import { YouTrackCtrl } from '../you-track/you-track-ctrl.js';

const MIN_STATE_CHANGE_DURATION_MS = 60 * 60 * 1000;
const DEFAULT_REMAINING_EFFORT_MS = 0;
const DEFAULT_WAIT_TIME_MS = 0;
/**
 * Resolution is 1 hour.
 */

const SCHEDULING_RESOLUTION_MS = 60 * 60 * 1000;
/**
 * Cannot preempt of split issues that are less than 4 hours ideal time.
 */

const MIN_ACTIVITY_DURATION = 4;
/**
 * Prefix for the IDs of external contributors.
 *
 * No YouTrack user ID must have the same prefix. Given that all
 * [YouTrack IDs are of form "x-y" where both x and y are integers](https://github.com/JetBrains/xodus/blob/master/entity-store/src/main/java/jetbrains/exodus/entitystore/PersistentEntityId.java)
 * (see also the
 * [`EntityId` interface](https://github.com/JetBrains/xodus/blob/master/openAPI/src/main/java/jetbrains/exodus/entitystore/EntityId.java)),
 * anything starting with a non-numeric character should already be safe.
 */

const EXTERNAL_CONTRIBUTOR_ID_PREFIX = '@fschopp/project-planning-ui-for-you-track/';
var Action;

(function (Action) {
  Action["CONNECT"] = "connect";
  Action["BUILD_PLAN"] = "build";
  Action["UPDATE_PREDICTION"] = "update";
  Action["STOP"] = "stop";
  Action["NOTHING"] = "nothing";
})(Action || (Action = {}));
/**
 * Controller for {@link App}.
 */


class AppCtrl {
  /**
   * Constructor.
   *
   * @param app the non-transient application state
   * @param appName the name of the application
   */
  constructor(app, appName = document.title) {
    this.app = app;
    this.appName = appName;
    /**
     * The alert controller for displaying alerts to the user.
     */

    this.alertCtrl = new AlertsCtrl();
    this.progress_ = S.value(undefined);
    this.extendedProjectPlan_ = S.value(undefined);
    this.connectSignal_ = S.data(null);
    this.youTrackMetadata_ = S.value(undefined);
    this.settingsCtrl = new SettingsCtrl(this.app.settings, this.youTrackMetadata_, this.connectSignal_);
    this.youTrackCtrl_ = new YouTrackCtrl(this.youTrackMetadata_, this.settingsCtrl.verifiedBaseUrl, this.alertCtrl);
    this.action = S(() => actionFromState(this.progress_(), this.youTrackCtrl_.pendingMetadata(), this.youTrackMetadata_(), this.extendedProjectPlan_(), toNormalizedPlainSettings(this.app.settings)));
    this.youTrackMetadata = this.youTrackMetadata_;
    this.progress = this.progress_;
    this.extendedProjectPlan = this.extendedProjectPlan_; // 'seed' is undefined (the calculation does not keep a state), and 'onchanges' is true (skip the initial run).

    S.on(this.connectSignal_, () => this.connect(), undefined, true); // Bind document.title to react to changes of this.app.transient.appName

    S(() => {
      const title = this.app.settings.name();
      document.title = title.length > 0 ? `${appName}: ${title}` : appName;
    });
  }

  showErrorIfFailure(title, promise) {
    promise.catch(exception => this.alertCtrl.alert(title, exception)).finally(() => {
      this.progress_(undefined);
    });
  }

  defaultAction() {
    const action = this.action();

    switch (action) {
      case Action.CONNECT:
        return this.connect();

      case Action.BUILD_PLAN:
        return this.showErrorIfFailure('Failed to build project plan', this.buildPlan());

      case Action.UPDATE_PREDICTION:
        return this.showErrorIfFailure('Failed to update prediction', this.updatePrediction());

      case Action.STOP:
      case Action.NOTHING:
        return;

      default:
        return unreachableCase(action);
    }
  }

  connect() {
    goToOauthPage(this.settingsCtrl.verifiedBaseUrl(), this.settingsCtrl.settings.youTrackServiceId(), this.app);
  }

  async buildPlan(currentConfig = toNormalizedPlainSettings(this.app.settings)) {
    const typeFieldId = currentConfig.typeFieldId;
    const splittableTypeIds = new Set(currentConfig.splittableTypeIds);
    const youTrackConfig = {
      stateFieldId: currentConfig.stateFieldId,
      inactiveStateIds: currentConfig.inactiveStateIds,
      remainingEffortFieldId: currentConfig.remainingEffortFieldId,
      remainingWaitFieldId: currentConfig.remainingWaitFieldId,
      assigneeFieldId: currentConfig.assigneeFieldId,
      otherCustomFieldIds: [typeFieldId],
      dependsLinkTypeId: currentConfig.dependsLinkTypeId,
      doesInwardDependOnOutward: currentConfig.doesInwardDependOnOutward,
      savedQueryId: currentConfig.savedQuery,
      overlaySavedQueryId: currentConfig.overlaySavedQuery,
      minStateChangeDurationMs: MIN_STATE_CHANGE_DURATION_MS,
      defaultRemainingEffortMs: DEFAULT_REMAINING_EFFORT_MS,
      defaultWaitTimeMs: DEFAULT_WAIT_TIME_MS,
      isSplittableFn: issue => {
        const value = issue.customFields[typeFieldId];
        return splittableTypeIds.has(value);
      }
    };
    this.reconstructProjectPlanResult = await reconstructProjectPlan(this.settingsCtrl.verifiedBaseUrl(), youTrackConfig, percentDone => this.progress_(percentDone));
    this.reconstructProjectPlanTimestamp = Date.now();
    await this.updatePrediction(currentConfig);
  }

  async updatePrediction(currentConfig = toNormalizedPlainSettings(this.app.settings)) {
    const youTrackMetadata = this.youTrackMetadata_();
    const contributors = [];
    const idToExternalContributorName = new Map();

    for (const contributor of currentConfig.contributors) {
      let id;
      let numMembers;

      if (contributor.type === 1
      /* EXTERNAL */
      ) {
          id = `${EXTERNAL_CONTRIBUTOR_ID_PREFIX}${idToExternalContributorName.size}`;
          idToExternalContributorName.set(id, contributor.name);
          numMembers = contributor.numMembers;
        } else {
        id = contributor.id;
        numMembers = 1;
      }

      contributors.push({
        id,
        minutesPerWeek: 60 * contributor.hoursPerWeek,
        numMembers
      });
    }

    const schedulingOptions = {
      contributors,
      minutesPerWeek: youTrackMetadata.minutesPerWorkWeek,
      resolutionMs: SCHEDULING_RESOLUTION_MS,
      minActivityDuration: MIN_ACTIVITY_DURATION,
      predictionStartTimeMs: this.reconstructProjectPlanTimestamp
    };
    const schedule = await scheduleUnresolved(this.reconstructProjectPlanResult.issues, schedulingOptions);
    const finalPlan = appendSchedule(this.reconstructProjectPlanResult, schedule, schedulingOptions.predictionStartTimeMs);

    if (isFailure(finalPlan)) {
      throw finalPlan;
    } else {
      this.extendedProjectPlan_({
        plan: finalPlan,
        settings: currentConfig,
        youTrackTimestamp: this.reconstructProjectPlanTimestamp,
        idToExternalContributorName
      });
    }
  }

}

function actionFromState(progress, pendingMetadata, youTrackMetadata, extendedProjectPlan, currentSettings) {
  if (progress !== undefined || pendingMetadata) {
    return Action.STOP;
  } else if (youTrackMetadata === undefined) {
    return Action.CONNECT;
  } else if (extendedProjectPlan === undefined) {
    return Action.BUILD_PLAN;
  }

  const newWithoutContributors = Object.assign({}, currentSettings);
  const lastWithoutContributors = Object.assign({}, extendedProjectPlan.settings);
  delete newWithoutContributors.contributors;
  delete lastWithoutContributors.contributors;
  const youTrackConfigChanged = !deepEquals(newWithoutContributors, lastWithoutContributors);
  const contributorsChanged = !deepEquals(currentSettings.contributors, extendedProjectPlan.settings.contributors);

  if (youTrackConfigChanged) {
    return Action.BUILD_PLAN;
  } else if (contributorsChanged) {
    return Action.UPDATE_PREDICTION;
  } else {
    return Action.NOTHING;
  }
}

export { Action, AppCtrl };
//# sourceMappingURL=app-ctrl.js.map
