import { handlePotentialOauthRedirect } from '@fschopp/project-planning-for-you-track';
import S from 's-js';
import { assignSettings } from '../settings/settings-model.js';
import { assignApp, Page } from './app-model.js';

const FABRICATED_HOST = 'http://ignored-host';

function isParsedHash(hash) {
  return typeof hash !== 'string';
}

function parseWindowLocation() {
  const theHash = window.location.hash;

  if (theHash.startsWith('#/')) {
    const fabricatedUrl = new URL(theHash.slice(1), 'http://ignored-host/');
    return {
      path: fabricatedUrl.pathname,
      params: fabricatedUrl.searchParams,
      raw: theHash
    };
  } else {
    return theHash;
  }
}
/**
 * Router that two-way binds the fragment (also called hash) of the URI of the current window/tab and the application
 * state.
 */


class Router {
  /**
   * Constructor.
   *
   * @param app the non-transient application state
   */
  constructor(app) {
    this.hash = S.value(parseWindowLocation());
    const savedState = handlePotentialOauthRedirect();

    if (savedState !== undefined) {
      assignApp(app, savedState);
    } // Bind hash to react to changes of window.location.hash


    const setStateFromHash = () => this.hash(parseWindowLocation());

    window.addEventListener('hashchange', setStateFromHash, false);
    S.cleanup(() => window.removeEventListener('hashchange', setStateFromHash)); // Bind config to react to changes of hash.

    const needsNormalization = S.data(null);
    let hashFromConfig = '';
    S(() => {
      const currentHash = this.hash();

      if (isParsedHash(currentHash)) {
        if (currentHash.raw.slice(1) === hashFromConfig) {
          // If this event is the result of us setting the hash ourselves (see below), ignore this update.
          return;
        }

        const candidatePageName = currentHash.path.slice(1);

        if (Object.values(Page).includes(candidatePageName)) {
          app.currentPage(candidatePageName);
        }

        if (currentHash.params.has('config')) {
          const config = JSON.parse(currentHash.params.get('config'));
          assignSettings(app.settings, config);
        } else {
          needsNormalization(null);
        }
      }
    }); // Bind browser hash to react to changes of config

    S(() => {
      // We need to depend on the following data signal, in order to guarantee normalization of the hash if the user
      // changes it manually. (We cannot use S.on(), which may be a bit clearer, because we also depend on the
      // configuration signals.)
      needsNormalization();
      const currentHash = window.location.hash.slice(1);
      const url = new URL(`/${app.currentPage()}`, FABRICATED_HOST); // JSON.stringify visits the object tree recursively, so even changes to nested data signal will trigger this
      // computation to re-run. What is remarkable: nested data signals (say in a SDataArray) that do not yet exist will
      // also trigger this computation in the future! The reason it works: The set of data signals a computation depends
      // on is not static but updated whenever the computation runs. Now adding an element to SDataArray reruns this
      // computation, which then records the dependency on the element's data signal(s). The same works when removing
      // elements.

      url.searchParams.set('config', JSON.stringify(app.settings));
      hashFromConfig = url.toString().substring(FABRICATED_HOST.length);

      if (currentHash !== hashFromConfig) {
        window.location.hash = hashFromConfig;
      }
    });
  }

  static create(app, assignSavedState) {
    const savedState = handlePotentialOauthRedirect();

    if (savedState !== undefined) {
      assignSavedState(savedState);
    }

    return new Router(app);
  }

}

export { Router };
//# sourceMappingURL=router.js.map
