import S from 's-js';
import { hrefFrom } from '../../utils/surplus.js';
import { ContributorsCtrl } from '../contributors/contributors-ctrl.js';

/**
 * Controller for settings related to project planning (YouTrack custom fields, list of available resources, etc.).
 */

class SettingsCtrl {
  /**
   * Constructor.
   *
   * @param settings
   * @param youTrackMetadata
   * @param connectSignal
   */
  constructor(settings, youTrackMetadata, connectSignal) {
    this.settings = settings;
    this.youTrackMetadata = youTrackMetadata;
    this.connectSignal = connectSignal;
    this.verifiedBaseUrl = S(() => {
      try {
        const youTrackUrlString = this.settings.youTrackBaseUrl();
        const url = new URL(youTrackUrlString);

        if (url.pathname.length === 0 || url.pathname.charAt(url.pathname.length - 1) !== '/') {
          url.pathname = url.pathname.concat('/');
        }

        return url.toString();
      } catch (exception) {
        if (!(exception instanceof TypeError)) {
          throw exception;
        }

        return '';
      }
    });
    this.metadata = S(() => {
      const youTrackMetadata = this.youTrackMetadata();
      return youTrackMetadata !== undefined && this.verifiedBaseUrl() === youTrackMetadata.baseUrl ? youTrackMetadata : emptyYouTrackMetadata();
    });
    this.enumFields = this.mapOfCustomFieldTypes('enum[1]');
    this.periodFields = this.mapOfCustomFieldTypes('period');
    this.stateFields = this.mapOfCustomFieldTypes('state[1]');
    this.userFields = this.mapOfCustomFieldTypes('user[1]');
    this.states = S(() => {
      const customField = this.stateFields().get(this.settings.stateFieldId());
      const array = customField === undefined ? [] : customField.fieldDefaults.bundle.values.filter(stateBundleElement => !stateBundleElement.isResolved).sort((left, right) => left.ordinal - right.ordinal);
      return array.reduce((map, element) => map.set(element.id, element), new Map());
    });
    this.types = S(() => {
      const customField = this.enumFields().get(this.settings.typeFieldId());
      const array = customField === undefined ? [] : customField.fieldDefaults.bundle.values.sort((left, right) => left.ordinal - right.ordinal);
      return array.reduce((map, element) => map.set(element.id, element), new Map());
    });
    this.directedIssueLinkTypes = S(() => this.metadata().issueLinkTypes.filter(linkType => linkType.directed).reduce((map, linkType) => map.set(linkType.id, linkType), new Map()));
    this.issueLinkDirectionNames = S(() => {
      const issueLinkType = this.directedIssueLinkTypes().get(this.settings.dependsLinkTypeId());
      return issueLinkType === undefined ? ['→', '←'] : [issueLinkType.sourceToTarget, issueLinkType.targetToSource];
    });
    this.savedQueryIds = S(() => new Set(this.metadata().savedSearches.map(savedQuery => savedQuery.id)));
    const youTrackUserMap = S(() => this.metadata().users.reduce((map, user) => map.set(user.id, user), new Map()));
    this.contributorsCtrl = new ContributorsCtrl(settings.contributors, youTrackUserMap); // If saved search and overlay are the same, set the overlay to none.

    S(() => {
      if (this.settings.savedQuery() === this.settings.overlaySavedQuery()) {
        this.settings.overlaySavedQuery('');
      }
    });
  }

  static currentUri() {
    const uri = new URL(window.location.href);
    uri.hash = '';
    uri.username = '';
    uri.password = '';
    uri.search = '';
    return uri.toString();
  }

  hrefRelativeToBaseUrl(relativePath) {
    const signal = S(() => {
      const baseUrl = this.verifiedBaseUrl();
      return baseUrl.length === 0 ? '' : new URL(relativePath, baseUrl).toString();
    });
    return hrefFrom(signal);
  }

  hubRelativeToBaseUrlAndServiceId(relativePath) {
    const signal = S(() => {
      const baseUrl = this.verifiedBaseUrl();
      const serviceId = this.settings.youTrackServiceId();
      return baseUrl.length === 0 || serviceId.length === 0 ? '' : new URL(relativePath(serviceId), this.verifiedBaseUrl()).toString();
    });
    return hrefFrom(signal);
  }

  mapOfCustomFieldTypes(fieldTypeId) {
    return S(() => this.metadata().customFields.filter(customField => customField.fieldType.id === fieldTypeId).reduce((map, customField) => map.set(customField.id, customField), new Map()));
  }

}

function emptyYouTrackMetadata() {
  return {
    baseUrl: '',
    customFields: [],
    issueLinkTypes: [],
    savedSearches: [],
    users: [],
    minutesPerWorkWeek: 5 * 8 * 60
  };
}

export { SettingsCtrl };
//# sourceMappingURL=settings-ctrl.js.map
