import { jsonable, toPlain } from '../../utils/s.js';
import S from 's-js';
import { createContributors, assignContributors } from '../contributors/contributors-model.js';

/**
 * Creates a new state for the settings UI component.
 */

function createSettings() {
  return {
    name: jsonable(S.value('')),
    youTrackBaseUrl: jsonable(S.value('')),
    youTrackServiceId: jsonable(S.value('')),
    stateFieldId: jsonable(S.value('')),
    inactiveStateIds: jsonable(S.value(new Set())),
    remainingEffortFieldId: jsonable(S.value('')),
    remainingWaitFieldId: jsonable(S.value('')),
    assigneeFieldId: jsonable(S.value('')),
    typeFieldId: jsonable(S.value('')),
    splittableTypeIds: jsonable(S.value(new Set())),
    dependsLinkTypeId: jsonable(S.value('')),
    doesInwardDependOnOutward: jsonable(S.value(true)),
    savedQuery: jsonable(S.value('')),
    overlaySavedQuery: jsonable(S.value('')),
    contributors: createContributors()
  };
}
/**
 * Updates the state of the settings UI component to the given values in a plain JSON object.
 *
 * The update is performed within a single S.js transaction.
 *
 * @param settings state of the settings UI component
 * @param plain plain JSON object
 */

function assignSettings(settings, plain) {
  S.freeze(() => {
    settings.name(plain.name);
    settings.youTrackBaseUrl(plain.youTrackBaseUrl);
    settings.youTrackServiceId(plain.youTrackServiceId);
    settings.stateFieldId(plain.stateFieldId);
    settings.inactiveStateIds(new Set(plain.inactiveStateIds));
    settings.remainingEffortFieldId(plain.remainingEffortFieldId);
    settings.remainingWaitFieldId(plain.remainingWaitFieldId);
    settings.assigneeFieldId(plain.assigneeFieldId);
    settings.typeFieldId(plain.typeFieldId);
    settings.splittableTypeIds(new Set(plain.splittableTypeIds));
    settings.dependsLinkTypeId(plain.dependsLinkTypeId);
    settings.doesInwardDependOnOutward(plain.doesInwardDependOnOutward);
    settings.savedQuery(plain.savedQuery);
    settings.overlaySavedQuery(plain.overlaySavedQuery);
    assignContributors(settings.contributors, plain.contributors);
  });
}
/**
 * Creates a new normalized plain JSON value for the given settings.
 *
 * Normalization means that the YouTrack base URL is the result of {@link normalizedBaseUrl}().
 */

function toNormalizedPlainSettings(settings) {
  const plainSettings = toPlain(settings);
  return Object.assign({}, plainSettings, {
    youTrackBaseUrl: normalizedBaseUrl(plainSettings.youTrackBaseUrl)
  });
}
/**
 * Returns the normalized YouTrack base URL.
 *
 * Normalization means returning a syntactically valid URL that ends with a slash (/).
 *
 * @return the normalized URL, or the empty string if the given URL is not valid
 */

function normalizedBaseUrl(baseUrl) {
  try {
    const url = new URL(baseUrl);

    if (url.pathname.length === 0 || url.pathname.charAt(url.pathname.length - 1) !== '/') {
      url.pathname = url.pathname.concat('/');
    }

    return url.toString();
  } catch (exception) {
    if (!(exception instanceof TypeError)) {
      throw exception;
    }

    return '';
  }
}

export { assignSettings, createSettings, normalizedBaseUrl, toNormalizedPlainSettings };
//# sourceMappingURL=settings-model.js.map
